/* locksim.h */

#ifndef LOCKSIM_H
#define LOCKSIM_H

enum token_type {
    TRANSACTION = 258,
    READ = 259,
    WRITE = 260,
    ID = 261,
    NUMBER = 262,
    DONE = 263,
    LOCK = 264,
    READLOCK = 265,
    WRITELOCK = 266,
    UNLOCK = 267
};

struct Symbol {
    char *name;
    int token_type;
};

struct Variable {
    int symbol_index;
    int value;
};

struct Environment {
    struct Variable *variables;
    int nr_variables;
    int variable_table_length;
};

struct TreeNode {
    int type;
    union {
        struct TreeNode *args[2];
        int leaf_value;
    };
};

enum transaction_state {
    RUNNABLE,
    BLOCKED,
    FINISHED
};

struct Transaction {
    struct TreeNode *name; /* A symbol node */
    struct TreeNode *program; /* The list of commands */
    struct TreeNode *rest_of_program; /* Remaining list of commands */
    struct Environment *environment;
    enum transaction_state state;
    int blocking_variable; /* If failed to lock, this is the symbol index of the database variable that we failed to lock */

};

enum lock_type {
    UNLOCKED,
    SHARED,
    EXCLUSIVE
};

struct Lock {
    enum lock_type type;
    int symbol_index;
    struct Transaction *transaction;
};

struct Database {
    struct Environment *environment;
    struct Lock *locks;
    int nr_locks;
    int lock_table_length;
};

extern void fatal_error(char *format, ...);

extern int lookup_or_insert_symbol(char *name, int token_type);
extern char *get_symbol_name_unsafe(int symbol_index);
extern int get_symbol_type(int symbol_index);
extern void init_symbols(void);

extern struct Environment *create_environment(void);
extern void destroy_environment(struct Environment *e);
extern int get_variable_value(struct Environment *environment, int symbol_index);
extern void set_variable_value(struct Environment *environment, int symbol_index, int value);
extern void show_environment(struct Environment *environment);
extern void show_environment_compact(struct Environment *environment);
extern int equal_environments(struct Environment *e1, struct Environment *e2);

extern struct TreeNode *parse_scenario(FILE *input);

extern struct Database *create_database(void);
extern void destroy_database(struct Database *d);
extern enum lock_type find_my_lock(struct Database *database, int symbol_index, struct Transaction *transaction);
extern int lock_variable(struct Database *database, int symbol_index, struct Transaction *transaction, enum lock_type wanted_type);
extern void show_locks(struct Database *database);

extern int scan(FILE *f);
extern int lexical_value;
extern int line_number;

extern struct TreeNode *parse_scenario(FILE *input);

extern struct TreeNode *create_node(int type, struct TreeNode *arg1, struct TreeNode *arg2);
extern struct TreeNode *create_number_node(int value);
extern struct TreeNode *create_id_node(int symbol_index);
extern struct TreeNode *car(struct TreeNode *cons_node);
extern struct TreeNode *cdr(struct TreeNode *cons_node);
extern void rplacd(struct TreeNode *cons_node, struct TreeNode *new_cdr);
extern int list_length(struct TreeNode *cons_node);
extern void print_scenario(struct TreeNode *tree);
extern void print_compact(struct TreeNode *p);

extern void register_result(struct Environment *environment, int was_deadlock);
extern void show_results(void);

extern struct Transaction *create_transaction(struct TreeNode *name, struct TreeNode *program);
extern void destroy_transaction(struct Transaction *t);
extern void show_transaction(struct Transaction *transaction);

extern void add_transaction(struct Transaction *t);
extern void clear_scheduler(void);
extern void block_transaction(struct Transaction *t, int blocking_variable);
extern void finish_transaction(struct Transaction *t);
extern void wakeup_others_waiting_for(int blocking_variable);
extern struct Transaction *get_random_runnable_transaction(void);
extern struct Transaction *find_transaction_by_name(int symbol_index);
extern int get_nr_blocked_transactions(void);
extern void show_transactions(void);
extern void show_transactions_compact(void);
extern void show_scheduler_state(void);

extern void execute_scenario(struct TreeNode *scenario, int verbosity);

extern int verbosity;

#endif
